package org.kaoriha.jspbp;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadBase;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.StringUtils;

class Request {
	static final String CONTINUE_ID_QN = "continuationId";
	static final String PERSISTENT_ID_QN = "persistentId";
	private static final String CONTINUE_OBJ_PROP_NAME = "continuation";
	private static final String SERVLET_CONTEXT_PROP_NAME = "servletContext";
	private static final int ID_LENGTH = 10;
	private final HttpServletRequest request;
	private final DelegateResponse response;
	private final HttpServletResponse origResponse;
	private final FilterChain chain;
	private final String behindPackageName;
	private final JspbpFilter filter;
	private Object behind;
	private Map<String, String[]> parameterMap;
	private Map<String, List<FileItem>> fileItemMap;

	Request(HttpServletRequest request, HttpServletResponse response,
			FilterChain chain, JspbpFilter filter) {
		this.request = request;
		this.response = new DelegateResponse(response, request.getRequestURI());
		origResponse = response;
		this.chain = chain;
		this.behindPackageName = filter.getBehindPackageName();
		this.filter = filter;
	}

	void doFilter() throws IOException, ServletException {
		try {
			doFilterImpl();
		} catch (IOException e) {
			throw e;
		} catch (ServletException e) {
			throw e;
		} catch (RedirectException e) {
			response.sendRedirect(request.getContextPath() + e.getLocation());
		} catch (ClassNotFoundException e) {
			chain.doFilter(request, origResponse);
		} catch (Exception e) {
			throw new ServletException(e);
		}
	}

	private void methodNotAllowed() throws IOException, ServletException {
		if (behind == null) {
			try {
				generateBehind();
			} catch (Exception e) {
				throw new ServletException(e);
			}
		}

		StringBuilder sb = new StringBuilder();
		if (behind instanceof BeforeGettable) {
			sb.append("GET");
			sb.append(", HEAD");
		}
		if (behind instanceof BeforePostable) {
			if (sb.length() > 0) {
				sb.append(", ");
			}
			sb.append("POST");
		}

		response.setHeader("Allow", sb.toString());
		response.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
	}

	@SuppressWarnings( { "deprecation", "unchecked" })
	private void buildParameterMap() throws ServletException,
			UnsupportedEncodingException, RedirectException {
		parameterMap = new HashMap<String, String[]>(request.getParameterMap());
		if (ServletFileUpload.isMultipartContent(request)) {
			fileItemMap = new HashMap<String, List<FileItem>>();
			DiskFileItemFactory factory = new DiskFileItemFactory();
			factory.setSizeThreshold(8000);
			if (filter.getTempDir() != null) {
				factory.setRepository(new File(filter.getTempDir()));
			}
			ServletFileUpload upload = new ServletFileUpload(factory);
			upload.setSizeMax(filter.getMaxFilesize());
			List items;
			try {
				items = upload.parseRequest(request);
			} catch (FileUploadBase.SizeLimitExceededException e) {
				if (filter.getSizeExceedPath() == null) {
					throw new ServletException(e);
				}
				throw new RedirectException(filter.getSizeExceedPath());
			} catch (FileUploadException e) {
				throw new ServletException(e);
			}

			for (Object o : items) {
				FileItem item = (FileItem) o;
				String k = item.getFieldName();
				if (item.isFormField()) {
					String v = item.getString(filter.getEncoding());
					if (parameterMap.containsKey(k)) {
						List<String> l = new ArrayList<String>(Arrays
								.asList(parameterMap.get(k)));
						l.add(v);
						parameterMap.put(k, l.toArray(new String[0]));
					} else {
						String[] ss = { v };
						parameterMap.put(k, ss);
					}
				} else if (item.getName() != null && item.getName().length() > 0) {
					if (fileItemMap.containsKey(k)) {
						fileItemMap.get(k).add(item);
					} else {
						List<FileItem> fil = new ArrayList<FileItem>();
						fil.add(item);
						fileItemMap.put(k, fil);
					}
				}
			}
		}
	}

	private String getParameter(String name) throws ServletException, RedirectException {
		if (parameterMap == null) {
			try {
				buildParameterMap();
			} catch (UnsupportedEncodingException e) {
				throw new ServletException(e);
			}
		}
		if (parameterMap.containsKey(name)) {
			return parameterMap.get(name)[0];
		}
		return null;
	}

	private void doFilterImpl() throws IOException, ServletException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException, NoSuchMethodException,
			RedirectException, ClassNotFoundException {
		String sp = request.getServletPath();
		if (sp.indexOf(".jsp") != sp.length() - 4) {
			throw new ServletException(
					"JSP's file extension must be \"jsp\". Otherwise filter-mapping in web.xml is wrong.");
		}

		String method = request.getMethod();
		if (!(method.equals("GET") || method.equals("POST") || method
				.equals("HEAD"))) {
			methodNotAllowed();
			return;
		}

		String persistentId = getParameter(PERSISTENT_ID_QN);
		if (persistentId == null) {
			generateBehind();
			if (behind instanceof Persistent) {
				String location;
				String q = PERSISTENT_ID_QN + "="
						+ RandomStringUtils.randomAlphanumeric(ID_LENGTH);
				String orig = request.getQueryString();
				if (orig == null || orig.length() == 0) {
					location = request.getRequestURI() + "?" + q;
				} else {
					location = request.getRequestURI() + "?" + orig + "&" + q;
				}
				response.sendRedirect(location);
				return;
			}
		} else {
			Object p = request.getSession().getAttribute(persistentId);
			if (p == null) {
				generateBehind();
				response.setPersistentId(persistentId);
			} else if (!p.getClass().getCanonicalName().equals(
							getBehindClassName())) {
				response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
				return;
			} else {
				behind = p;
				response.setPersistentId(persistentId);
			}
		}
		copyProperty();

		String continueId = getParameter(CONTINUE_ID_QN);
		if (continueId != null) {
			Object continueObj = request.getSession().getAttribute(continueId);
			if (continueObj != null
					&& PropertyUtils
							.isWriteable(behind, CONTINUE_OBJ_PROP_NAME)) {
				try {
					PropertyUtils.setSimpleProperty(behind, CONTINUE_OBJ_PROP_NAME,
							continueObj);
				} catch (IllegalArgumentException e) {
					// bad continueId
					PropertyUtils.setSimpleProperty(behind, CONTINUE_OBJ_PROP_NAME,
							null);
				}
			}
		}

		if (method.equals("GET") || method.equals("HEAD")) {
			beforeGet();
		} else {
			beforePost();
			if (persistentId != null) {
				request.getSession().setAttribute(persistentId, behind);
			}
		}

		if (!response.isRedirected() && !response.isError()) {
			if (!method.equals("POST")) {
				response.setHeader("Cache-Control",
						"no-store, no-cache, must-revalidate");
				response.setHeader("Pragma", "no-cache");
				response.setHeader("Expires", "-1");
			}

			request.setAttribute("behind", behind);
			chain.doFilter(request, origResponse);
		}
	}

	private String getBehindClassName() {
		String sp = request.getServletPath();
		StringBuilder className = new StringBuilder(sp.substring(0,
				sp.length() - 4)); // remove ".jsp"

		while (true) {
			int i = className.indexOf("/");
			if (i == -1) {
				break;
			}
			className.setCharAt(i, '.');
		}

		// capitalize
		int fci = className.lastIndexOf(".") + 1;
		className.setCharAt(fci, Character.toUpperCase(className.charAt(fci)));

		className.insert(0, behindPackageName);

		return className.toString();
	}

	private void generateBehind() throws InstantiationException,
			IllegalAccessException, ClassNotFoundException {
		behind = Class.forName(getBehindClassName()).newInstance();
	}

	@SuppressWarnings( { "unchecked", "deprecation" })
	private void copyProperty() throws IllegalAccessException,
			InvocationTargetException, NoSuchMethodException, ServletException {
		if (PropertyUtils.isWriteable(behind, SERVLET_CONTEXT_PROP_NAME)) {
			PropertyUtils.setSimpleProperty(behind, SERVLET_CONTEXT_PROP_NAME, filter.getServletContext());
		}

		if (ServletFileUpload.isMultipartContent(request)) {
			for (Map.Entry<String, List<FileItem>> entry : fileItemMap
					.entrySet()) {
				List<FileItem> vl = entry.getValue();
				switch (vl.size()) {
				case 0:
					break;
				case 1:
					if (PropertyUtils.isWriteable(behind, entry.getKey())) {
						PropertyUtils.setSimpleProperty(behind, entry.getKey(),
								vl.get(0));
					}
					break;
				default:
					if (PropertyUtils.isWriteable(behind, entry.getKey()
							+ "List")) {
						PropertyUtils.setSimpleProperty(behind, entry.getKey()
								+ "List", vl);
					}
					break;
				}
			}
		}

		
		Set<String> checkboxNameSet = new HashSet<String>();
		if (parameterMap.containsKey(Constant.CHECKBOX_NAME_QN)) {
			String[] v = { "false" };
			for (String cn : parameterMap.get(Constant.CHECKBOX_NAME_QN)) {
				if (!parameterMap.containsKey(cn)) {
					parameterMap.put(cn, v);
				}
				checkboxNameSet.add(cn);
			}
		}

		for (Map.Entry<String, String[]> entry : parameterMap.entrySet()) {
			String k = entry.getKey();
			String[] va = entry.getValue();
			if (checkboxNameSet.contains(k)) {
				if (PropertyUtils.isWriteable(behind, k)) {
					PropertyUtils.setSimpleProperty(behind, k, va[0]
							.equals("true"));
				}
				continue;
			}

			switch (va.length) {
			case 0:
				break;
			case 1:
				if (PropertyUtils.isWriteable(behind, k)) {
					if (StringUtils.isEmpty(va[0])) {
						PropertyUtils.setSimpleProperty(behind, k, null);
					} else {
						PropertyUtils.setSimpleProperty(behind, k, va[0]);
					}
				}
				break;
			default:
				if (PropertyUtils.isWriteable(behind, k + "List")) {
					PropertyUtils.setSimpleProperty(behind, k + "List", Arrays
							.asList(va));
				}
				break;
			}
		}
	}

	private void beforeGet() throws IOException, ServletException {
		if (!(behind instanceof BeforeGettable)) {
			methodNotAllowed();
			return;
		}
		BeforeGettable gettable = (BeforeGettable) behind;
		gettable.beforeGet(request, response);
	}

	private void beforePost() throws IOException, ServletException,
			IllegalAccessException, InvocationTargetException,
			NoSuchMethodException {
		if (!(behind instanceof BeforePostable)) {
			methodNotAllowed();
			return;
		}
		BeforePostable postable = (BeforePostable) behind;

		if (PropertyUtils.isReadable(behind, CONTINUE_OBJ_PROP_NAME)) {
			String cid = RandomStringUtils.randomAlphanumeric(ID_LENGTH);
			response.setContinueId(cid);

			postable.beforePost(request, response);

			if (response.isRedirected()) {
				Object cobj = PropertyUtils.getSimpleProperty(behind,
						CONTINUE_OBJ_PROP_NAME);
				if (cobj != null) {
					request.getSession().setAttribute(cid, cobj);
				}
			}
		} else {
			postable.beforePost(request, response);
		}
	}
}

class RedirectException extends Exception {
	private String location;

	public RedirectException(String location) {
		this.location = location;
	}

	public String getLocation() {
		return location;
	}
}